# Re-export all symbols from submodules for backward compatibility
from dagster._symbol_annotations.annotatable import (
    Annotatable as Annotatable,
    T_Annotatable as T_Annotatable,
    copy_annotations as copy_annotations,
)
from dagster._symbol_annotations.lifecycle import (
    BetaInfo as BetaInfo,
    DeprecatedInfo as DeprecatedInfo,
    PreviewInfo as PreviewInfo,
    SupersededInfo as SupersededInfo,
    attach_deprecation_info_and_wrap as attach_deprecation_info_and_wrap,
    beta as beta,
    beta_param as beta_param,
    deprecated as deprecated,
    deprecated_param as deprecated_param,
    get_beta_info as get_beta_info,
    get_beta_param_info as get_beta_param_info,
    get_beta_params as get_beta_params,
    get_deprecated_info as get_deprecated_info,
    get_deprecated_param_info as get_deprecated_param_info,
    get_deprecated_params as get_deprecated_params,
    get_preview_info as get_preview_info,
    get_superseded_info as get_superseded_info,
    has_beta_params as has_beta_params,
    has_deprecated_params as has_deprecated_params,
    hidden_param as hidden_param,
    is_beta as is_beta,
    is_beta_param as is_beta_param,
    is_deprecated as is_deprecated,
    is_deprecated_param as is_deprecated_param,
    is_preview as is_preview,
    is_superseded as is_superseded,
    only_allow_hidden_params_in_kwargs as only_allow_hidden_params_in_kwargs,
    preview as preview,
    superseded as superseded,
)
from dagster._symbol_annotations.public import (
    PUBLIC as PUBLIC,
    PublicAttr as PublicAttr,
    is_public as is_public,
    public as public,
)
